/*
** Module   : ScreenDefinition.java
** Abstract : contains the configuration of a given frame or family of frames
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050705   @21680 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050713   @21692 Redesign according to the new design
**                           approach.
** 003 EVL 20050715   @21705 Changing the location of the widget types
**                           constants.
** 004 EVL 20050718   @21798 Adding the method to add new created config
**                           to the content of the screen definition. This
**                           will be used when the screen definition is
**                           created on the server side.
** 005 EVL 20050720   @21799 Changes to meet the new XML formats. Each
**                           screen definition holds only one frame, root
**                           or child. So all widgets inside this screen
**                           definition will be the child widgets. Also
**                           the ID variable is added to the class to
**                           identify the parent-child relationship
**                           between the frames if exists.
** 006 EVL 20050727   @21859 Change the method getWidgetsInFrame().
**                           The child frame must also be excluded from
**                           the enumeration process.
** 007 EVL 20050827   @22373 Adding out of the range checking for method
**                           to return the config for the given widget ID.
** 008 SIY 20050831   @22545 Formatting and spelling fixes.
** 009 EVL 20051115   @23336 Moving the code to set up widget's frameID to
**                           the addConfig() method body.
** 010 SIY 20060209   @24450 Fixed Javadoc.
** 011 GES 20060802   @28414 Transmission performance improvements by
**                           shifting to be an externalizable. Many
**                           cleanups to meet coding standards.
** 012 SIY 20060824   @28856 Removed references to DialogBox.
** 013 SIY 20071121   @35930 Support for Frame 'need pause' flag 
**                           maintenance. Some refactoring.
**                           Ref: #35928
** 014 SIY 20080715   @39113 Pause flag is changed to more general 
**                               'suppress redraw' flag.
** 015 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 016 EVL 20140822          Rework for new widget ID usage approach.  No more incline ID
**                           calculations.  Adding the dynamic widget support to be able to remove
**                           the config to dynamic widget.
** 017 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 018 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 019 VIG 20150116          Added menuId and Id is renamed to frameId.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.io.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.client.*;

/**
 * This is a container for the configuration of a given frame or family of
 * frames which is efficiently serialized to the client from the server.
 * 
 * This container can also be used for Menu configuration. In that case an {@link #menuId}
 * will contain root menu id and {@link id} will have its default value.
 */
public class ScreenDefinition
implements Externalizable
{
   /** The widget configuration data. */
   private LinkedHashMap<Integer, WidgetConfig> widgets =  new LinkedHashMap<>();

   /** Root frame ID. */
   private int frameId = 0;
   
   /** Root menu ID. */
   private int menuId = 0;
   
   /** Suppress redraw on push of screen definition. */
   private boolean suppressRedraw = false; 
   
   /** Map with the registered side-labels in this frame. */
   private Map<Integer, WidgetConfig> sideLabels = new HashMap<>();
   
   /** For each side-label, specify its parent. */
   private Map<Integer, Integer> sideLabelParents = new HashMap<>();
   
   /**
    * Default constructor (only used for serialization).
    */
   public ScreenDefinition()
   {
   }
   
   /**
    * The constructor creates a screen definition with the given root frame
    * ID.
    * 
    * @param    id
    *           Identifies the root frame (or the only frame if this is not
    *           going to hold the configuration of a frame family).
    */
   public ScreenDefinition(int id)
   {
      this(id, false);
   }
   
   /**
    * The constructor creates a screen definition with the given root frame
    * or menu ID.
    * 
    * @param    id
    *           Identifies the root frame/menu (or the only frame if this is not
    *           going to hold the configuration of a frame family).
    * @param    menu
    *           <code>true</code> if this definition will be used for menu. 
    */
   public ScreenDefinition(int id, boolean menu)
   {
      if (menu)
      {
         this.menuId = id;
      }
      else
      {
         this.frameId = id;
      }
   }
   
   /**
    * Get the root frame ID.
    * 
    * @return   The root frame ID.
    */
   public int getRootFrameID()
   {
      return frameId;
   }
   
   /**
    * Get the root menu ID.
    * 
    * @return   The root menu ID.
    */
   public int getRootMenuID()
   {
      return menuId;
   }
   
   /**
    * Gets an array of all widget IDs in a given frame including any child
    * frames.
    * 
    * @param    frameId
    *           ID of the frame for which widget IDs are requested.
    * 
    * @return   The list of all widget IDs for the specified frame.
    */
   public ArrayList<Integer> getWidgetsInFrame(int frameId)
   {
      ArrayList<Integer> wids = new ArrayList<Integer>();
      Iterator<WidgetConfig> iter = getConfigsIterator();
 
      while (iter.hasNext())
      {
         WidgetConfig config = iter.next();
         
         if (config.frameId == frameId && !(config instanceof FrameConfig))
         {
            wids.add(new Integer(config.id.getId()));
         }
      }

      return wids;
   }
   
   /**
    * Gets an array of all widget IDs in a given menu including any child
    * frames.
    * 
    * @param    menuId
    *           ID of the menu for which widget IDs are requested.
    * 
    * @return   The list of all widget IDs for the specified menu.
    */
   public ArrayList<Integer> getWidgetsInMenu(int menuId)
   {
      ArrayList<Integer> wids = new ArrayList<Integer>();
      Iterator<WidgetConfig> iter = getConfigsIterator();
 
      while (iter.hasNext())
      {
         WidgetConfig config = iter.next();
         
         if (!(config instanceof MenuElementConfig))
            continue;
         
         if (((MenuElementConfig)config).parentId() == menuId)
         {
            wids.add(new Integer(config.id.getId()));
         }
      }

      return wids;
   }
   
   /**
    * Get an iterator to enumerate all widget configurations inside this
    * screen definition.
    * 
    * @return   The iterator.
    */
   public Iterator<WidgetConfig> getConfigsIterator()
   {
      return widgets.values().iterator();
   }

   /**
    * Get an iterator to enumerate all widget configurations inside this
    * screen definition.
    * 
    * @return   The iterator.
    */
   public Iterator<WidgetConfig> getSideLabelsIterator()
   {
      return sideLabels.values().iterator();
   }
   
   /**
    * Gets a specific widget configuration instance for the widget with the
    * ID provided.
    * 
    * @param    id
    *           The widget ID.
    * 
    * @return   The requested widget configuration object if exists,
    *           <code>null</code> otherwise.
    */
   public WidgetConfig getConfig(int id)
   {
      return widgets.get(new Integer(id));
   }
   
   /**
    * Add a widget configuration to the current screen definition.
    * 
    * @param    config
    *           New config to be added to the current screen definition.
    */
   public void addConfig(WidgetConfig config)
   {
      addConfigInt(config);
   }

   /**
    * Add a widget configuration (to the current screen definition) and
    * assign that widget the specified ID.
    * 
    * @param    config
    *           New config to be added to the current screen definition.
    * @param    widgetID
    *           Widget ID.
    */
   public void addConfig(WidgetConfig config, int widgetID)
   {
      config.id = new WidgetId(widgetID);
      addConfigInt(config);
   }
   
   /**
    * Removes a widget configuration from the current screen definition.
    * 
    * @param    config
    *           The config to be removed from the current screen definition.
    */
   public void removeConfig(WidgetConfig config)
   {
      int wid = config.id == null ? -1 : config.id.getId();
      
      if (widgets.get(wid) == null)
      {
         throw new RuntimeException("Missing widget for remove, ID = " + wid);
      }
      
      widgets.remove(wid);
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      frameId        = in.readInt();
      menuId         = in.readInt();    
      suppressRedraw = in.readBoolean();
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), widgets);
      WidgetEntry.arrayToMap((WidgetEntry[]) in.readObject(), sideLabels);
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeInt(frameId);
      out.writeInt(menuId);
      out.writeBoolean(suppressRedraw);
      out.writeObject(WidgetEntry.mapToArray(widgets));
      out.writeObject(WidgetEntry.mapToArray(sideLabels));
   }   
   
   /**
    * Set new value for 'suppress redraw' flag.
    * 
    * @param   suppressRedraw
    *          New value of flag.
    */
   public void setSuppressRedraw(boolean suppressRedraw)
   {
      this.suppressRedraw = suppressRedraw; 
   }

   /**
    * Get current value of 'suppress redraw' flag.
    * 
    * @return  The current value of the flag. 
    */
   public boolean isSuppressRedraw()
   {
      return suppressRedraw; 
   }

   /**
    * Add a new side-label and associate it with the given widget.
    * 
    * @param   sideLabelId
    *          The label's ID.
    * @param   widgetId
    *          The ID of the widget to which this label belongs.
    */
   public void addSideLabel(int sideLabelId, int widgetId)
   {
      if (sideLabelId < 0 || this.sideLabels.containsKey(sideLabelId))
      {
         // if is not for a valid server-side widget or is already registered, nothing to do
         return;
      }
      
      ConfigManager mgr = ConfigManager.getInstance();
      WidgetConfig lblCfg = mgr.getActiveConfig(new WidgetId(sideLabelId));
      
      if (lblCfg == null)
      {
         throw new NullPointerException(
            "No active config is registered for side-label with id " + sideLabelId + "!");
      }
      
      this.sideLabels.put(sideLabelId, lblCfg);
      
      this.sideLabelParents.put(sideLabelId, widgetId);
   }
   
   /**
    * Check if the given widget ID is for a side-label.
    * 
    * @param    widgetId
    *           The widget ID to be checked.
    *           
    * @return   <code>true</code> if the {@link #sideLabels} maps contains an entry for this 
    *           widget ID.
    */
   public boolean isSideLabel(int widgetId)
   {
      return sideLabels.containsKey(widgetId);
   }
   
   /**
    * Get the parent widget for the given side-label.
    * 
    * @param    sideLabelId
    *           The ID of the side label.
    *           
    * @return   The parent widget ID, as contained in the {@link #sideLabelParents} map.
    */
   public int getSideLabelParent(int sideLabelId)
   {
      return sideLabelParents.get(sideLabelId);
   }

   /**
    * Common worker to add a widget configuration to the screen definition.
    * 
    * @param    config
    *           Widget configuration.
    */
   private void addConfigInt(WidgetConfig config)
   {
      if (config.frameId == -1)
      {
         if (!(config instanceof FrameConfig))
         {
            config.frameId = frameId;
         }
      }
      
      int wid = (config.id == null ? -1 : config.id.getId());
      
      if (wid == -1)
      {
         assert (!SessionManager.get().isLeaf());
         
         // allocate an ID
         // we must never be here on the client side
         config.id = WidgetId.nextID();
         wid = config.id.getId();
      }
      
      Integer iwid = new Integer(wid);
      
      if (widgets.get(iwid) != null)
      {
         throw new RuntimeException("Duplicate widget ID = " + iwid);
      }
      
      widgets.put(iwid, config);  
   }
}
